/*
  * Copyright (c) 2018 Graham Payne
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, 
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITscroll_Y, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
  Morphing Clock: Author - Graham Payne, August 2018
  ESP-12E NodeMCU module, RTC clock module, DHT11 temp/humidity module, 64x32 P3 LED matrix panel
  Displays morphing digital clock, day, date, month
  Displays Analog clock, day, date, month
  displays upward scrolling weather from OpenWeathermap.org and local temp/humidity from DHT11 sensor.
  Has a kitchen timer function
  NTP server time update with time zone selecter
  OTA (over the air) update of firmware
  clock is controlled by Blynk app on phone thru WiFi to Blynk local server on RPi 3
  
  Thanks to:
  Hari Wiguna for concept and morphing code
  Dominic Buchstaller for PxMatrix
  lmirel MorphingClockRemix for further inspiration
  and countless others

  MUST USE version 5.13.2 ArduinoJson library
*/

/*  PIN TRANSLATIONS FOR NODEMCU ESP 12E***********
    D3 = GPIO 0
    D10 = GPIO 1
    D4 = GPIO 2
    D9 = GPIO 3
    D2 = GPIO 4
    D1 = GPIO 5
    D6 = GPIO 12
    D7 = GPIO 13
    D5 = GPIO 14
    D8 = GPIO 15
    D0 = GPIO 16
*/

#include <BlynkSimpleEsp8266.h>
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <Wire.h>
#include <TimeLib.h>
#include <RTClib.h>
#include <DHT.h>
#include <Ticker.h>
#include <PxMatrix.h>
#include <Fonts/Org_01.h>

// Local includes
#include "Digit.h"              // morphing digits

Ticker display_ticker;       // make Ticker instance for display updating

#define double_buffer
#define PI 3.1415926535897932384626433832795

// Pins for LED MATRIX..   (see schematic diagram for other connectiions on the matrix panel)
#define P_LAT 16        //D0
#define P_A       3        //D9
#define P_B       1        //D10
#define P_C     15        //D8
#define P_D     12        //D6
#define P_OE    2        //D4
PxMATRIX display(64, 32, P_LAT, P_OE, P_A, P_B, P_C, P_D);   // make an instance of the Matrix display, E not required.

// RTC config
#define  RTC_SCL             5                       // RTC - SCL chip select  -  D1 GPIO5
#define  RTC_SDA             4                       // RTC - SDA clock        -  D2 GPIO4
RTC_DS1307 RTC;                                    // make instance of the Real Time Clock

// DHT11 temp & humidity config
#define DHTTYPE DHT11                          // DHT uses DHT 11
#define DHTPIN                  0                      // DHT uses pin GPIO0 D3     
DHT dht(DHTPIN, DHTTYPE);                  //make an instance of the DHT11

// Morphing clock digits 
Digit digit0(&display, 0, 63 - 1 - 9 * 1, 4, display.color565(0, 0, 255));
Digit digit1(&display, 0, 63 - 1 - 9 * 2, 4, display.color565(0, 0, 255));
Digit digit2(&display, 0, 63 - 4 - 9 * 3, 4, display.color565(0, 0, 255));
Digit digit3(&display, 0, 63 - 4 - 9 * 4, 4, display.color565(0, 0, 255));
Digit digit4(&display, 0, 63 - 7 - 9 * 5, 4, display.color565(0, 0, 255));
Digit digit5(&display, 0, 63 - 7 - 9 * 6, 4, display.color565(0, 0, 255));

// Color defines RGB
int red = display.color565 (255, 0, 0);
int orange = display.color565 (255, 110, 0);
int yellow = display.color565 (255, 255, 0);
int green = display.color565 (0, 255, 0);
int blue = display.color565 (0, 0, 255);
int magenta = display.color565 (255, 0, 255);
int cyan = display.color565 (0, 250, 150);
int white = display.color565 (255, 255, 255);
int blck = display.color565 (0, 0, 0);

// Arrays for months and days
String Months[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"}; //char*
String Days[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

 // saves for blanking
unsigned int prev_second, prev_Timer_second, prev_dayOfTheWeek, prev_day, prev_month, prev_year;   

//morphing clock variables
unsigned int s0, s1, m0, m1, h0, h1;        

int timeZone = 10;     //  ......QLD AU time is +10              ************ CHANGE TO SUIT TIME ZONE ***************

// weather  variables
String Wstr;
int scroll_Y, old_scroll_Y;
unsigned long last = 0;           // scroll timer variables
unsigned long cm;                  // scroll timer variables
int h, t, oldh, oldt;                    //indoor temp/humidity sensor
int temp = -10000;                  // openweather variables
int pressure = -10000;
int humidity = -10000;
int wind_speed = -10000;
int wind_deg = -10000;
String dir;                               // wind direction
unsigned long sun_rise, sun_set;
int condition = 0;  // 0 - unk, 1 - sunny, 2 - cloudy, 3 - overcast, 4 - rainy, 5 - thunders, 6 - snow
String apiKey   = "????????"; // e.g a hex string like "abcdef0123456789abcdef0123456789"   open weather map api key
String location = "??????"; //e.g. "Paris, FR"  the city you want the weather for ** SET THIS TO YOUR LOCAL WEATHER STATION *****************************************************
char server[]   = "api.openweathermap.org";

// timer variables
int timer_Seconds, timer_Minutes, timer_Hours;      

// date variables
String ddstr, mmstr, yystr, allstr;

// brightness
int cin = 30;                                              // LED intensity as a percentage

// Flags
bool RunDigital_Flag = true;
bool RunAnalog_Flag = false;
bool RunWeather_Flag = false;
bool first_time = true;                     // first time to weather weather flag after is every 5minutes 30 secs
bool RunTimer_Flag = false;

// variables for analog clock
int handY, handX, Analog_hour;     // point on circle for hands
int centerX =  15;      // center of clock
int centerY = 15;
int ssRad = 13;         // second hand radius
int mmRad = 11;       // minute hand radius
int hhRad = 8;           // hour hand radius
int hhMark = 15;        // hour markers radius
float handAngle;        // 0 deg points East, 90 deg point South, 180 deg point West, 270 deg point north

// Set password for network.
char ssid[] = "?????";                                                // local wifi network SSID
char pass[] = "?????";                                                      // local network password
char auth[] = "??????";           // Blynk authority token

 // weather client
WiFiClient client;                                                                    // make an instance of the weather client                   

// NTP Servers:
WiFiUDP Udp;                                                                        // make an instance of WiFUDP
static const char ntpServerName[] = "us.pool.ntp.org";
unsigned int localPort = 8888;                                               // local port to listen for UDP packets
const int NTP_PACKET_SIZE = 48;                                     // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE];                             //buffer to hold incoming & outgoing packets
time_t getNtpTime();
void sendNTPpacket(IPAddress &address);

WidgetTerminal terminal(V6);                                            //make instance of app 'terminal' widget, used for report NTP time setting & weather on app

//*******************************************************************************
// ISR for display refresh
void display_updater() {
  display.display(70);
}

//*******************************************************************************
// Blynk run the digital clock button
BLYNK_WRITE(V0) {
  Prep (0, param.asInt());
  display.setFont();
  RefreshDigitalTime();
}

//*******************************************************************************
// Blynk run analog clock button
BLYNK_WRITE(V1) {
  Prep (1, param.asInt());
  display.setFont();
  DrawAnalogDate();
}

//*******************************************************************************
// Blynk run weather button
BLYNK_WRITE(V2) {
  Prep (2, param.asInt());
  display.setFont(&Org_01);
   scroll_Y = 36;
   old_scroll_Y = 37;
   
}

//*******************************************************************************
// Blynk set the clock NTP time button
BLYNK_WRITE(V3) {
  Prep (3, 0);
  Udp.begin(localPort);
  terminal.println("Starting UDP Local Port: ");
  terminal.println(Udp.localPort());
  delay(200);
  terminal.println("waiting for sync");
  setSyncProvider(getNtpTime);
  setSyncInterval(300);
  RTC.adjust(DateTime(year(), month(), day(), hour(), minute(), second()));
  delay(4000);                                                              //delay time to read terminal
  display.setFont();
  terminal.clear();
  Prep (0, 1);
}

//*******************************************************************************
// Blynk Timer button
BLYNK_WRITE(V4) {
  Prep (4, param.asInt());
  display.setFont();
  display.setTextColor (green);
  display.setCursor (17, 0);
  display.print ("TIMER");
  Blynk.syncVirtual(V7, V8, V9);            // get values from server
  
  digit0.Draw(timer_Seconds % 10);     //draw clock 
  digit1.Draw(timer_Seconds / 10);
  digit2.Draw(timer_Minutes % 10);
  digit3.Draw(timer_Minutes / 10);
  digit4.Draw(timer_Hours % 10);
  digit5.Draw(timer_Hours / 10);
  
  RunTimer_Flag = true;                         // start the timer
}

//********************************************************************************
// set panel brightness from Blynk slider
BLYNK_WRITE(V5) {
  cin = param.asInt();
  Adjust_Intensity();
}

//********************************************************************************
// Blynk timer settings
BLYNK_WRITE(V7) {
  timer_Seconds = param.asInt();
}
BLYNK_WRITE(V8) {
  timer_Minutes = param.asInt();
}
BLYNK_WRITE(V9) {
  timer_Hours = param.asInt();
}

//********************************************************************************
// Blynk timezone setting
BLYNK_WRITE(V10) {
  timeZone = param.asInt();
}

//*****************************************************************************
// preperations when clocks, weather & timer buttons pressed
void Prep (int v, int par) {
  RunDigital_Flag = false;
  RunAnalog_Flag = false;
  RunWeather_Flag = false;
  RunTimer_Flag = false;
  Blynk.virtualWrite (V0, LOW);           //switch off 'digital'  button
  Blynk.virtualWrite (V1, LOW);           //switch off 'analog' button
  Blynk.virtualWrite (V2, LOW);           // switch off 'weather' button
  Blynk.virtualWrite (V3, LOW);           // switch off 'set time' button
  Blynk.virtualWrite (V4, LOW);           // switch off 'timer' button
        
  switch (v) {
    case 0:
       RunDigital_Flag = true;
       if (par == 1) {
            Blynk.virtualWrite (V0, HIGH); 
            terminal.clear();
       }
       break;
    case 1:
       RunAnalog_Flag = true;
       if (par == 1) {
            Blynk.virtualWrite (V1, HIGH); 
            terminal.clear();
       }
       break;
    case 2:
       RunWeather_Flag = true;
       if (par == 1) Blynk.virtualWrite (V2, HIGH); 
       first_time = true;                                              // for run weather first time
       break;          
   case 3:
       Blynk.virtualWrite (V3, HIGH);                         // NTP  flag done in button routine
       break;      
   case 4:
       if (par == 1) Blynk.virtualWrite (V4, HIGH);     // Timer flag done in button routine
       break;    
   }
  
  if (par == 0)  display_ticker.detach(); else  display_ticker.attach(0.002, display_updater);    //turn off display if button off, only blanks display if ALL buttons are off
  display.fillScreen (blck);      // blank the screen
  Adjust_Intensity();
}

//********************************************************************************
// display and run the timer
void Display_Timer() {
  DateTime now = RTC.now();                                                                 //used as a second timer only
  if (timer_Seconds != 0 || timer_Minutes != 0 || timer_Hours != 0) {        // if not timed out
    if (prev_Timer_second != now.second()) {                                           // and it's a new second
    
      display.fillRect(2, 13, 59, 15, blck);        
      DrawColons();                                             
      timer_Seconds -- ;
      if (timer_Seconds < 0 && timer_Minutes > 0 || timer_Seconds < 0 && timer_Hours > 0) {
        timer_Seconds = 59;
        timer_Minutes -- ;
      }
      if (timer_Minutes < 0  && timer_Hours == 0) timer_Minutes = 0;
      if (timer_Minutes < 0  && timer_Hours > 0) {
        timer_Minutes = 59;
        timer_Hours -- ;
      }
      digit0.Draw(timer_Seconds % 10);                                                     //digit0.Draw(timer_Seconds % 10);    draw clock without morphing
      digit1.Draw(timer_Seconds / 10);
      digit2.Draw(timer_Minutes % 10);
      digit3.Draw(timer_Minutes / 10);
      digit4.Draw(timer_Hours % 10);
      digit5.Draw(timer_Hours / 10);
    }                                                                                                           // end  if (prev_Timer_second
    prev_Timer_second = now.second();
  }                                                                                                             // end if (timer_Seconds
  else {
    display.fillRect(0, 0, 63, 8, blck);
    display.setTextColor (red);
    display.setCursor (2, 0);
    display.print ("TIMER DONE");
    Blynk.virtualWrite (V4, LOW);                                                           // switch off 'timer' button
    
  }
}

//*******************************************************************************
// Run display sequence of date, digital clock, temp & humidity
void DisplayDigital() {
  DateTime now = RTC.now();                              //RTC: dayOfTheWeek starts Sun = 0 etc, months: Jan = 1 etc
  // ************ DATE ***************
  if (now.day() != prev_day) {                                  // if the date changes 
    display.fillRect(0, 0, 63, 8, blck);                        // blank out the date
    allstr = String(Days [now.dayOfTheWeek()]) + " " + String(now.day()) + " " +  String(Months [now.month() - 1]);    //compose the date
    display.setTextColor (green);
    display.setCursor (2, 0);
    display.print (allstr);                                               // print the date
  }
  
  // *********** TIME**************************
  s0 = now.second() % 10;
  s1 = now.second() / 10;
  m0 = now.minute() % 10;
  m1 = now.minute() / 10;
  h0 = now.hour() % 10;
  h1 = now.hour() / 10;

  if (s0 != digit0.Value())digit0.Morph(s0);                //morph the time
  if (s1 != digit1.Value()) digit1.Morph(s1);
  if (m0 != digit2.Value()) digit2.Morph(m0);
  if (m1 != digit3.Value()) digit3.Morph(m1);
  if (h0 != digit4.Value()) digit4.Morph(h0);
  if (h1 != digit5.Value()) digit5.Morph(h1);

  //************************ SAVE ***************************************
  prev_dayOfTheWeek = now.dayOfTheWeek();     // save for next time
  prev_day = now.day();
  prev_month = now.month();


}

//*******************************************************************************
// Run display sequence of date, analog clock, temp & humidity
void DisplayAnalog() {
  DateTime now = RTC.now();
 
  if (now.day() != prev_day) {                                                 // date
      DrawAnalogDate();
  }
  
  if (now.second() != prev_second) {                                     // time
      display.fillCircle(centerX, centerY, ssRad + 1, blck);      // blank the analog clock area HANDS only. Note trying to overwrite the hands in black didn't work, remnants were left behind.
      handAngle = (now.second() * 6) - 90;                              //draw the second hand. Note angle 0 degree faces East
      DrawHand(centerX, centerY, ssRad,  handAngle, red);
      handAngle = (now.minute() * 6) - 90;                               //draw the minute hand
      DrawHand(centerX, centerY, mmRad,  handAngle, green);
      if (now.hour() > 11) Analog_hour = now.hour() - 12; else  Analog_hour = now.hour();         //set hour to 12 hour system, 0 - 11 for analog clock
      handAngle = (Analog_hour * 30) + ((now.minute() / 12) * 6) - 90;                                         //get to correct hand position, i.e.  2:24  is at 12 min mark to simulate real clock
      DrawHand(centerX, centerY, hhRad,  handAngle, blue);                                                      //draw the hour hand
    }
  
  prev_day = now.day();                                                            // save
  prev_second = now.second();
}

// ***********************************************************************
// Draw the analog hour marks
void drawHourMarks() {
  for ( int x = 0; x < 360; x += 30) {                                                //draw the hour markers every 30 degrees
    handY = round(centerY + hhMark * sin( x * (PI / 180) ) );
    handX = round(centerX + hhMark * cos( x * (PI / 180) ) );
    display.drawPixel (handX, handY, orange);
  }
}

// ***********************************************************************
// Draw the analog clock hands
void DrawHand( int cx, int cy, int r, int a, int c) {
  int x = round( cx + r * cos( a * (PI / 180) ) );
  int y = round( cy + r * sin( a * (PI / 180) ) );
  display.drawLine (cx, cy, x, y, c);
}

// ***********************************************************************
// Draw the analog clock date, day & month
void DrawAnalogDate() {
      DateTime now = RTC.now();
      display.fillRect(38, 0, 18, 32, blck);                   // if the date changes blank out the date
      allstr = String(Days [now.dayOfTheWeek()]);                                            // print day
      display.setTextColor (blue);
      display.setCursor (39, 0);
      display.print (allstr);
    
      allstr = String(Months [now.month() - 1]);                                                    // print month
      display.setCursor (39, 24);
      display.print (allstr);
    
      Digit digit0(&display, 0, 48, 9, display.color565(0, 0, 255 * cin / 100));        //print date
      Digit digit1(&display, 0, 39, 9, display.color565(0, 0, 255 * cin / 100));
      digit0.Draw(now.day() % 10);                                                                      //draw date without morphing
      digit1.Draw(now.day() / 10);
}

//*******************************************************************************
// Refresh the digital time after blanking
void RefreshDigitalTime() {
  display.fillScreen (blck);                      // blank the screen
  
  DateTime now = RTC.now();

  allstr = String(Days [now.dayOfTheWeek()]) + " " + String(now.day()) + " " +  String(Months [now.month() - 1]);    //compose the date
  display.setTextColor (green);
  display.setCursor (2, 0);
  display.print (allstr);                                               // print the date
  
  DrawColons();
  digit0.Draw(now.second() % 10);       //draw clock without morphing
  digit1.Draw(now.second() / 10);
  digit2.Draw(now.minute() % 10);
  digit3.Draw(now.minute() / 10);
  digit4.Draw(now.hour() % 10);
  digit5.Draw(now.hour() / 10);
}

//*******************************************************************************
// draw colons for digital clock
void DrawColons() {
  digit1.DrawColon(display.color565(0, 0, 255 * cin / 100));
  digit3.DrawColon(display.color565(0, 0, 255 * cin / 100));
}

//*****************************************************************************
// Adjust brightness/intensity of entire display
void Adjust_Intensity() {
  red = display.color565 (255 * cin / 100, 0, 0);
  orange = display.color565 (255 * cin / 100, 110 * cin / 100, 0);
  yellow = display.color565 (255 * cin / 100, 255 * cin / 100, 0);
  green = display.color565 (0, 255 * cin / 100, 0);
  blue = display.color565 (0, 0, 255 * cin / 100);
  magenta = display.color565 (255 * cin / 100, 0, 255 * cin / 100);
  cyan = display.color565 (0, 255 * cin / 100, 150 * cin / 100);
  white = display.color565 (255 * cin / 100, 255 * cin / 100, 255 * cin / 100);
  
if (RunDigital_Flag) {
      digit0.SetColor(display.color565(0, 0, 255 * cin / 100));
      digit1.SetColor(display.color565(0, 0, 255 * cin / 100));
      digit2.SetColor(display.color565(0, 0, 255 * cin / 100));
      digit3.SetColor(display.color565(0, 0, 255 * cin / 100));
      digit4.SetColor(display.color565(0, 0, 255 * cin / 100));
      digit5.SetColor(display.color565(0, 0, 255 * cin / 100));
      RefreshDigitalTime();
    }
  if (RunAnalog_Flag) {
    drawHourMarks();
    DrawAnalogDate();
  }
}

//*****************************************************************************
// for scrolling the weather text up the screen
void Scroll() {
    display.setTextColor (blck);
    display.setCursor (0, old_scroll_Y);
    display.print (Wstr);                                      // blank out Wstr on previous line
    
    display.setTextColor (green);
    display.setCursor (0, scroll_Y);
    display.print (Wstr);                                       // print Wstr on new line
}

// *********************************************************************************
// Setup
void setup() {
  ArduinoOTA.setHostname("MatrixClock");                                  //added for OTA
  ArduinoOTA.onError([](ota_error_t error) { ESP.restart(); });
  ArduinoOTA.begin();

  display.begin(16);
  display_ticker.attach(0.002, display_updater);   // timer interrupt for the display refresh
                                                                              //** everything marked ** can be deleted or commented for a faster boot up**********************
  display.setFont(&Org_01);                                 //** set font to 5x6
  display.setTextColor (blue);                                //**
  display.setCursor (4, 18);                                   //**
  display.println ("Matrix Clock");                           //** splash
  display.drawRect(1, 11, 62, 11, red);                 //**
  delay(5000);                                                        //** delay to admire splash
  
  display.fillScreen (blck);                                      //** blank the screen
  display.setCursor (0, 4);                                     //**
  display.println ("OTA begin");                              //**

  Wire.begin(RTC_SDA, RTC_SCL);                   // Starts the Wire library allows I2C communication to the Real Time Clock
  RTC.begin();                                                       // Starts communications to the RTC
 //RTC.adjust(DateTime(F(__DATE__), F(__TIME__)));     //sets RTC at compile time from PC...do only once required, use battery after
 										// ensure there is a battery fitted to the RTC module
  
  display.println ("Wire begin");                               //**
  display.println ("RTC begin");                               //**

  pinMode(DHTPIN, INPUT_PULLUP);
  dht.begin();
  display.println ("DHT begin");                               //**
  delay(3000);                                                         //**

  Blynk.begin(auth, ssid, pass, IPAddress(192, 168, 31, ???), 8080);    //RPi local server port 8080 ** CHANGE TO SUIT YOUR SERVER

  display.fillScreen (blck);                                       //** blank the screen
  display.setCursor (0, 4);                                      //**
  display.println ("Blynk v " + String(BLYNK_VERSION));     //**
  display.println("connected");                                //**
  display.println("IP address: ");                              //**
  display.println(WiFi.localIP());                              //**
  delay(3000);                                                         //**
  
  display.setFont();                                                // set to standard font
  Blynk.virtualWrite (V5, cin);                               //set panel brightness control in the app, 20 - 100 %
  Blynk.virtualWrite (V10, timeZone);                   // set timezone in the app
  Prep (0, 1);                                                         //prepare for digital clock
 
  terminal.println(F("Blynk v" BLYNK_VERSION ": Device started"));  // display Blynk library version on app
  terminal.println();
  terminal.flush();

  randomSeed(analogRead(5));                         // randomise for whatever purpose

}

// **************************************************************************************
// Loop
void loop() {
  Blynk.run();
  ArduinoOTA.handle();                                 //added for OTA
  if (RunDigital_Flag) DisplayDigital();
  if (RunAnalog_Flag) DisplayAnalog();
  if (RunTimer_Flag) Display_Timer();
  if (RunWeather_Flag) {
       DisplayWeather(); 
       cm = millis ();                    
       if ((cm - last)  > 200) {                            //scrolling timer
          last = cm;
          old_scroll_Y = scroll_Y;                      //scroll_Y starts at 36 (31 plus 7) at the bottom, allow 7 dots per line 
          scroll_Y --;
          if (scroll_Y < -56) scroll_Y = 36;         // chars are 7 high x 6 wide (incl spacing) i.e -14 is 2 rows to scroll off the top
          Scroll();
      }
  }
}

//*********************************************************************************
// Get time from NTP server
time_t getNtpTime() {
  IPAddress ntpServerIP;                                                  // NTP server's ip address
  while (Udp.parsePacket() > 0) ;                                      // discard any previously received packets
  terminal.println("Transmit NTP request");
  WiFi.hostByName(ntpServerName, ntpServerIP);         // get a random server from the pool
  terminal.println(ntpServerName);
  terminal.println(ntpServerIP);
  sendNTPpacket(ntpServerIP);
  uint32_t beginWait = millis();
  while (millis() - beginWait < 1500) {
    int size = Udp.parsePacket();
    if (size >= NTP_PACKET_SIZE) {
      Udp.read(packetBuffer, NTP_PACKET_SIZE);       // read packet into the buffer
      unsigned long secsSince1900;
      // convert four bytes starting at location 40 to a long integer
      secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
      secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
      secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
      secsSince1900 |= (unsigned long)packetBuffer[43];
      terminal.println("Received NTP response");
      terminal.println("DONE");
      terminal.println();
      terminal.flush();
      return secsSince1900 - 2208988800UL + timeZone * SECS_PER_HOUR;
    }
  }
  terminal.println();
  terminal.println("No NTP Response :-(");
  terminal.println("Try again");
  terminal.flush();
  delay(2000);
  return 0;                                                                        // return 0 if unable to get the time
}

//****************************************************************************
// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address) {
  memset(packetBuffer, 0, NTP_PACKET_SIZE);          // set all bytes in the buffer to 0
                                                                                         // Initialize values needed to form NTP request
                                                                                         // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;                                     // LI, Version, Mode
  packetBuffer[1] = 0;                                                       // Stratum, or type of clock
  packetBuffer[2] = 6;                                                       // Polling Interval
  packetBuffer[3] = 0xEC;                                                // Peer Clock Precision
                                                                                         // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
                                                                                        // all NTP fields have been given values, now
                                                                                        // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123);                                  //NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}

// *********************************************************************
// display weather
void DisplayWeather() {
  DateTime now = RTC.now();
  if (first_time) {                                                             // get the weather anyway
    first_time = false;
    getWeather ();
    draw_weather ();
  }
  if (now.second() == 30 && ((now.minute() % 5) == 0)) {     //updates every 10 minutes 
    getWeather ();
    draw_weather ();
  }
}

//*********************************************************************************
// convert wind direction to standard N, NNW,NW, W etc
String WindDirection(unsigned int wd) {
  if (wd >= 0 && wd <= 11) dir = "N";
  if (wd >= 12 && wd <= 34) dir = "NNE";
  if (wd >= 35 && wd <= 56) dir = "NE";
  if (wd >= 57 && wd <= 78) dir = "ENE";
  if (wd >= 79 && wd <= 101) dir = "E";
  if (wd >= 102 && wd <= 124) dir = "ESE";
  if (wd >= 125 && wd <= 146) dir = "SE";
  if (wd >= 147 && wd <= 168) dir = "SSE";
  if (wd >= 169 && wd <= 191) dir = "S";
  if (wd >= 192 && wd <= 204) dir = "SSW";
  if (wd >= 205 && wd <= 226) dir = "SW";
  if (wd >= 227 && wd <= 259) dir = "WSW";
  if (wd >= 260 && wd <= 281) dir = "W";
  if (wd >= 282 && wd <= 304) dir = "WNW";
  if (wd >= 305 && wd <= 326) dir = "NW";
  if (wd >= 327 && wd <= 348) dir = "NNW";
  if (wd >= 349 && wd <= 360) dir = "N";
  return dir;                                                           //return direction
}

//***********************************************************************
// get weather
void getWeather () {
  display.fillScreen(blck);
  display.setTextColor (red);
  display.setCursor (14, 14);
  display.println ("updating");
  display.setCursor (14, 20);
  display.print ("weather");

  if (!apiKey.length ()) {
    terminal.println("No Key");
    return;
  }
  terminal.println("connecting");
  if (client.connect (server, 80)) {
    terminal.println ("connected");
    client.print ("GET /data/2.5/weather?");                            // Make a HTTP request:
    terminal.println ("GET /data/2.5/weather?");
    client.print ("q=" + location);
    terminal.println ("q=" + location);
    client.print ("&appid=" + apiKey);
    terminal.println ("&appid=" + apiKey);
    client.print ("&cnt=3");
    terminal.println ("&cnt=3");
    client.println ("&units=metric");
    terminal.println ("&units=metric");
    client.println ("Host: api.openweathermap.org");
    terminal.println ("Host: api.openweathermap.org");
    client.println ("Connection: close");
    terminal.println ("Connection: close");
    client.println ();
  }
  else
  {
    terminal.println ("cant connect");
    return;
  }
  delay (1000);

  String line = "";
  String sval = "";
  int bT, bT2;
  while (client.connected ()) {
    if (client.available()) line = client.readStringUntil('\n');
    if (!line.length ()) terminal.println ("can't retrieve weather data");
    else
    {
      terminal.println ("weather");
      terminal.println (line);
      bT = line.indexOf ("\"main\":\"");                        //weather conditions - find "main":"?????",
      if (bT > 0) {
        bT2 = line.indexOf ("\",\"", bT + 8);
        sval = line.substring (bT + 8, bT2);
        terminal.println ("condition " + String(sval));
        if (sval.equals("Clear")) condition = 1;            //0 - unk, 1 - sunny, 2 - cloudy, 3 - overcast, 4 - rainy, 5 - thunders, 6 - snow
        else if (sval.equals("Clouds")) condition = 2;
        else if (sval.equals("Overcast")) condition = 3;
        else if (sval.equals("Rain")) condition = 4;
        else if (sval.equals("Thunderstorm")) condition = 5;
        else if (sval.equals("Snow")) condition = 6;
      }

      bT = line.indexOf ("\"temp\":");                        // find temperature
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 7);
        sval = line.substring (bT + 7, bT2);
        terminal.println ("Temperature " + String(sval));
        temp = sval.toInt ();
      }
      else terminal.println ("no temperature");

      bT = line.indexOf ("\"pressure\":");                    // find pressure
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 11);
        sval = line.substring (bT + 11, bT2);
        terminal.println ("pressure " + String(sval));
        pressure = sval.toInt();
      }
      else terminal.println ("No Pressure");

      bT = line.indexOf ("\"humidity\":");                    // find humidity
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 11);
        sval = line.substring (bT + 11, bT2);
        terminal.println ("humidity " + String(sval));
        humidity = sval.toInt();
      }
      else terminal.println ("No Humidity");

      bT = line.indexOf ("\"speed\":");                    //find wind speed
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 8);
        sval = line.substring (bT + 8, bT2);
        terminal.println ("wind speed " + String(sval));
        wind_speed = sval.toInt() * 3.6;                 // convert mtrs/second to kph
      }
      else terminal.println ("No speed");

      bT = line.indexOf ("\"deg\":");                       // find wind direction
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 6);
        sval = line.substring (bT + 6, bT2 - 1);
        terminal.println ("wind direction " + String(sval));
        wind_deg = sval.toInt();
      }
      else terminal.println ("No wind direction");

      bT = line.indexOf ("\"sunrise\":");                    // find sunrise
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 10);
        sval = line.substring (bT + 10, bT2);
        sun_rise = sval.toInt();                                 // UNIX time UDT
        sun_rise += (timeZone * 3600);                  // add our time zone as seconds
        if(minute(sun_rise) < 10) {                          // using timelib function to convert UNIX time to hours and minutes
            terminal.println("Sunrise " + String(hour(sun_rise)) + ":0" + String(minute(sun_rise))); 
        } else {
            terminal.println("Sunrise " + String(hour(sun_rise)) + ":" + String(minute(sun_rise))); 
        }
    }
      else terminal.println ("No sunrise");

      bT = line.indexOf ("\"sunset\":");                    // sunset
      if (bT > 0) {
        bT2 = line.indexOf (",\"", bT + 9);
        sval = line.substring (bT + 9, bT2);
        sun_set = sval.toInt();                                  // UNIX time UDT
        sun_set += (timeZone * 3600);                   // add our time zone as seconds
        if(minute(sun_set) <10) {
             terminal.println("Sunset " + String(hour(sun_set)) + ":0" + String(minute(sun_set)));  // display as hours:minutes
        } else {
             terminal.println("Sunset " + String(hour(sun_set)) + ":" + String(minute(sun_set)));
        }
    }
      else terminal.println ("No sunset");

      terminal.println();
      terminal.println();
      terminal.flush();
    }                                                                   // if line length
  }                                                                     // while connected
}

//***********************************************************************
void draw_weather () {
  display.fillScreen(blck);

  if (temp == -10000 || humidity == -10000 || pressure == -10000) {       // no data received
    display.fillScreen(blck);
    display.setTextColor (red);
    display.setCursor (10, 18);
    display.print ("no data");
    delay(1000);
    return;
  } else {

   // ******************** indoor temp/humidty line 1  **************************************************************
    h = 0;
    t = 0;
    h = dht.readHumidity();
    t = dht.readTemperature();                                                      // or dht.readTemperature(true) for Fahrenheit
    if (h == -1 || t == -1) {                                                            // these sensors are unstable, report -1 sometimes
        h = oldh;                                                                               // if no reading use last readings
        t = oldt;
    } else {
        oldh = h;
        oldt = t;
  }
  
    Wstr = String(t) + "'C inside          ";                                        // indoor temp/humidity  line 1
    Wstr += String(temp) + "'C outside           ";                           // outdoor temperature  line 2
    Wstr += String(h) + "% inside             ";                                  // indoor humidity  line 3
    Wstr += String (humidity) + "% outside                ";                // outdoor humidity  line 4
    Wstr += String (pressure)  + "mb press            ";                  // pressure line 5 
    WindDirection (wind_deg);                                                     // convert degrees to Wind Direction
    Wstr += dir + " at " + String (wind_speed) + "kph           ";     // direction & wind speed lline 6

    switch (condition) {                                                                  // get conditions line 7
      case 1: Wstr += String("Clear                  "); break;
      case 2: Wstr += String("Cloudy                 "); break;
      case 3: Wstr += String("Overcast             "); break;
      case 4: Wstr += String("Rainy                   "); break;
      case 5: Wstr += String("Stormy                "); break;
      case 6: Wstr += String("Snowy                 "); break;
    }

     if(minute(sun_rise) < 10) {
             Wstr += String(hour(sun_rise)) + ":0" + String(minute(sun_rise)) + " sunrise         ";      //sunrise line 8
        } else {
            Wstr += String(hour(sun_rise)) + ":" + String(minute(sun_rise)) + " sunrise         ";         //sunrise line 8
        }

    if(minute(sun_set) < 10) {
            Wstr += String(hour(sun_set)) + ":0" + String(minute(sun_set)) + " sunset          ";         //sunset line 9
       } else {
            Wstr += String(hour(sun_set)) + ":" + String(minute(sun_set)) + " sunset          ";           //sunset line 9
       }
   }
}


// +++++++++++++++++++ INFO ON LIBRARIES ++++++++++++++++++++++++++++
/* **********PxMatrix Functions **************************
  begin(uint8_t row_pattern);
  begin();
  clearDisplay(void);
  display(uint16_t show_time);   // Updates the display
  drawPixelRGB565(int16_t x, int16_t y, uint16_t color);                                                                   // Draw pixels
  drawPixelRGB565(int16_t x, int16_t y, uint16_t color, bool selected_buffer);
  drawPixel(int16_t x, int16_t y, uint16_t color);
  drawPixel(int16_t x, int16_t y, uint16_t color, bool selected_buffer);
  drawPixelRGB888(int16_t x, int16_t y, uint8_t r, uint8_t g,uint8_t b);
  drawPixelRGB888(int16_t x, int16_t y, uint8_t r, uint8_t g,uint8_t b, bool selected_buffer);
  displayTestPattern(uint16_t showtime);                                                                                          // Helpful for debugging (place in display update loop)
  displayTestPixel(uint16_t show_time);                                                                                           // Helpful for debugging (place in display update loop)
  flushDisplay();                                                                                                                                 // FLush the buffer of the display
  setRotate(bool rotate);                                                                                                                     // Rotate display
  setFastUpdate(bool fast_update);                                                                                                    // Helps to reduce display update latency on larger displays
  setPanelsWidth(uint8_t panels);                                                                                                     // Set the number of panels that make up the display area width
*/

/*
 * ******* Digit Functions ******************************
   Digit(PxMATRIX* d, byte value, uint16_t xo, uint16_t yo, uint16_t color);
    void Draw(byte value);
    void Morph(byte newValue);
    void DrawColon(uint16_t c);
    void SetColor(uint16_t c);
*/


// **************************** Adafruit GFX DRAWING FUNCTIONS**************************

/*  drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint16_t color) {
    drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color)
    drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color)
    drawCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color)
    fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color),
    fillScreen(uint16_t color),
    drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
    drawCircleHelper(int16_t x0, int16_t y0, int16_t r, uint8_t cornername, uint16_t color),
    fillCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color),
    fillCircleHelper(int16_t x0, int16_t y0, int16_t r, uint8_t cornername, int16_t delta, uint16_t color),
    drawTriangle(int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color),
    fillTriangle(int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color),
    drawRoundRect(int16_t x0, int16_t y0, int16_t w, int16_t h, int16_t radius, uint16_t color),
    fillRoundRect(int16_t x0, int16_t y0, int16_t w, int16_t h, int16_t radius, uint16_t color),
    drawBitmap(int16_t x, int16_t y, const uint8_t bitmap[], int16_t w, int16_t h, uint16_t color),
    drawBitmap(int16_t x, int16_t y, const uint8_t bitmap[], int16_t w, int16_t h, uint16_t color, uint16_t bg),
    drawBitmap(int16_t x, int16_t y, uint8_t *bitmap, int16_t w, int16_t h, uint16_t color),
    drawBitmap(int16_t x, int16_t y, uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, uint16_t bg),
    drawXBitmap(int16_t x, int16_t y, const uint8_t bitmap[], int16_t w, int16_t h, uint16_t color),
    drawGrayscaleBitmap(int16_t x, int16_t y, const uint8_t bitmap[], int16_t w, int16_t h),
    drawGrayscaleBitmap(int16_t x, int16_t y, uint8_t *bitmap, int16_t w, int16_t h),
    drawGrayscaleBitmap(int16_t x, int16_t y, const uint8_t bitmap[], const uint8_t mask[], int16_t w, int16_t h),
    drawGrayscaleBitmap(int16_t x, int16_t y, uint8_t *bitmap, uint8_t *mask, int16_t w, int16_t h),
    drawRGBBitmap(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h),
    drawRGBBitmap(int16_t x, int16_t y, uint16_t *bitmap, int16_t w, int16_t h),
    drawRGBBitmap(int16_t x, int16_t y, const uint16_t bitmap[], const uint8_t mask[], int16_t w, int16_t h),
    drawRGBBitmap(int16_t x, int16_t y, uint16_t *bitmap, uint8_t *mask, int16_t w, int16_t h),
    drawChar(int16_t x, int16_t y, unsigned char c, uint16_t color, uint16_t bg, uint8_t size),
    setCursor(int16_t x, int16_t y),
    setTextColor(uint16_t c),
    setTextColor(uint16_t c, uint16_t bg),
    setTextSize(uint8_t s),
    setTextWrap(boolean w),
    cp437(boolean x=true),
    setFont(const GFXfont *f = NULL),
    getTextBounds(char *string, int16_t x, int16_t y, int16_t *x1, int16_t *y1, uint16_t *w, uint16_t *h),
*/    getTextBounds(const __FlashStringHelper *s, int16_t x, int16_t y, int16_t *x1, int16_t *y1, uint16_t *w, uint16_t *h);
